@@ -444,8 +444,7 @@ static void write_inode_tables(ext2_filsys fs, int lazy_flag, int itable_zeroed)
 				      _("done                            \n"));
 
 	/* Reserved inodes must always have correct checksums */
-	if (fs->super->s_creator_os == EXT2_OS_LINUX &&
-	    ext2fs_has_feature_metadata_csum(fs->super))
+	if (ext2fs_has_feature_metadata_csum(fs->super))
 		write_reserved_inodes(fs);
 }
 
@@ -1699,6 +1698,11 @@ profile_error:
 			break;
 		case 'L':
 			volume_label = optarg;
+			if (strlen(volume_label) > EXT2_LABEL_LEN) {
+				volume_label[EXT2_LABEL_LEN] = '\0';
+				fprintf(stderr, _("Warning: label too long; will be truncated to '%s'\n\n"),
+					volume_label);
+			}
 			break;
 		case 'm':
 			reserved_ratio = strtod(optarg, &tmp);
@@ -2696,10 +2700,21 @@ static void fix_cluster_bg_counts(ext2_filsys fs)
 static int create_quota_inodes(ext2_filsys fs)
 {
 	quota_ctx_t qctx;
+	errcode_t retval;
 
-	quota_init_context(&qctx, fs, QUOTA_ALL_BIT);
+	retval = quota_init_context(&qctx, fs, QUOTA_ALL_BIT);
+	if (retval) {
+		com_err(program_name, retval,
+			_("while initializing quota context"));
+		exit(1);
+	}
 	quota_compute_usage(qctx);
-	quota_write_inode(qctx, quotatype_bits);
+	retval = quota_write_inode(qctx, quotatype_bits);
+	if (retval) {
+		com_err(program_name, retval,
+			_("while writing quota inodes"));
+		exit(1);
+	}
 	quota_release_context(&qctx);
 
 	return 0;
